#!/usr/bin/perl
#
# exiscan 0.99 by tom kistner <tom@duncanthrax.net>
#
# Thanks to
# 	- Marcus Klein <marcus@netline-ag.de> (AVP support)
#	- Chris Beauchamp <chris@buildstore.co.uk> (Sophos Support)
#	- Rick Ballard <Richard.Ballard@xwave.com> (some good suggestions ;)
#	- Jordi Bosveld <jotti@jotti.dhs.org> (RAV antivirus support)
#	- Dave Hill <dhill@zeus.com> (Scanner output support)
#	- Sergey Kovalyov <sergey@metaphoria.net> (alternative (demo) AVP support)
#
# URL: http://duncanthrax.net/exiscan/
#
# !!! see INSTALL for installation instructions first !!!
# !!!  configuration has been moved to 'exiscan.cf'   !!!
#

$version = "0.99";

use Unix::Syslog qw(:macros);
use Unix::Syslog qw(:subs); 
use File::Copy;
use File::Remove qw(remove);

require "@ARGV[0]";

# chdir to base path, so core files will go into the right directory <g>
chdir("$basepath");

# now some sanity checks

print "Checking for mail .. ";
if (!(-x $mail)) {
	print "not found ($mail) ! Please check your settings\n";
	exit(1);
};
print "found\n";


print "Checking for reformime .. ";
if (!(-x $reformime)) {
	print "not found. Falling back to metamail.\n";
	$reformime = 0;
	print "Checking for metamail .. ";
	if (!(-x $metamail)) {
		print "not found ($metamail) ! Please check your settings\n";
		exit(1);
	};
	print "found\n";
}
else {
	print "found, using reformime to unpack MIME mails.\n";
};

print "Checking for metasend .. ";
if (!(-x $metasend)) {
        print "not found ($metasend) ! Please check your settings\n";
        exit(1);
};
print "found\n";

print "Checking for tnef .. ";
if (!(-x $tnef)) {
        print "not found, will not unpack MS-TNEF format\n";
        $tnef = 0;
} 
else {
	print "found\n";
};

print "Checking for virus scanner .. ";
if (!(-x $scannerex)) {
        print "not found ($scannerex) ! Please check your settings\n";
        exit(1);
};
print "found\n";

print "Checking for exim .. ";
if (!(-x $exim)) {
        print "not found ($exim) ! Please check your settings\n";
        exit(1);
};
print "found\n";

$quit = 0;
$| = 0;

# fork a child and exit, if debug mode is not enabled

if (!($debug)) {
	FORK: {
		if ($pid = fork) {
			print "exiscan launched into the background (pid $pid)\n";
			exit(0);
		}
		elsif (defined $pid) {
			goto CHILD;
		}
		else {
			die "exiscan can't fork to the background: $!\n";
		};
	};
} 
else {
	print "DEBUG mode enabled, running on terminal, CTRL-C to abort\n";
};

CHILD:

%checked = ();
$lastreset = time();

# ---------------------------------
# the exit handler
# HUP signal makes exiscan quit
# gracefully

sub catch_zap {
       my $signame = shift;
       $quit = 1;        
}

$SIG{HUP} = \&catch_zap;

# --------------------------------------
# set up syslog

use Unix::Syslog qw(:macros);
use Unix::Syslog qw(:subs);

openlog "exiscan", LOG_PID, $facility;

syslog LOG_INFO, "started";


# --------------------------------------
# the superloop
# it is in an eval block to trap unexpected errors.
# the handler will log an error with the mail ID
# and restart the superloop
# setting $quit to 1 (via HUP signal) is the only 
# way to exit this superloop correctly ...

SUPERLOOP:
eval { 
print "Entered superloop.\n" if ($debug);
while(!($quit)) {

# --------------------------------------
# forget scanned mails IDs every interval
# to allow re-dequeueing

 if ((time() - $lastreset) > $resetinterval) {
	 print "ID cache cleared.\n" if ($debug);
 	 syslog LOG_INFO, " exiscan clearing ID cache\n";
	 $lastreset = time();
	 %checked = ();
 };

# --------------------------------------
# get the contents of the queue (headers)

 opendir(QUEUEDIR,"$queuedir");
 @mails = grep /\-H/, readdir QUEUEDIR;
 closedir(QUEUEDIR);
 print "Scanned $queuedir and found this: @mails\n" if ($debug);

# --------------------------------------
# once second grace to decrease load
# and allow exim to perhaps finish
# writing a header file

 sleep 1;

# ---------------------------------------
# process all found mails

 foreach $mheader (@mails) {

	if ($quit) { goto EXITEXISCAN };

	$id = $mheader; $id =~ s/\-H$//g;

	print "Entering process loop for $id\n" if ($debug);

	if ( (-e "$queuedir/$id-H") && (!($checked{$id})) ) {

               # -------------------------------------
               # check for X-Scanner in the header

		open(HEADER,"< $queuedir/$id-H");
		if ((@scanner = grep /082  X-Scanner\: exiscan/, <HEADER>)) {
			
			close(HEADER);

			print "$id: header read\n" if ($debug);

		       # ------------------------------------
                       # see if WE did really scan it
		       # we don't trust anyone else ...
			
			($rest,$oldid,$cryptid,$rest2) = split /\*/, @scanner[$#scanner], 4;
			$checkid = crypt($oldid,$salt);
			$checkid =~ s/^..//;

			if ($checkid eq $cryptid) {

				if ($checked{$oldid}) {
					print "$id: resent, prev ID $oldid\n" if ($debug);
					syslog LOG_INFO, "ID $id R:resent, prev ID $oldid, dequeuing";
				}
				else {
					print "$id: delayed, attempting dequeue again\n" if ($debug);				
					syslog LOG_INFO, "ID $id R:delayed, attempting dequeue again";
				};
			
				$checked{$id} = 1;
			
                       	       # ------------------------------------
                               # de-queue 
                      
				print "$id: launching exim dequeue process\n" if ($debug);
				system("$exim -Mc $id &");


                               # ------------------------------------
                               # process next mail

				print "$id: leaving process loop\n" if ($debug);
				next;
			
			}
			else {
				print "$id: last X-Scanner line not from us\n" if ($debug);
				syslog LOG_INFO, "ID $id R: last X-Scanner line is not from us. Scanning now.";
			};
		}
		else {
			print "$id: this is a new one\n" if ($debug);
			close(HEADER);
		};

	       # -------------------------------------
	       # build complete mail in standard format
	       # get real TO and FROM addresses
	       
		# void the variables, just for paranoia
		@completemail = ();
		$from = "<EMPTY>";
		@to = ();
		$i = "";
		$subject = "<EMPTY>";
		
		open(HEADER,"< $queuedir/$id-H");
		do {
			if ($i =~ /^\</) { $from = $i; chop($from); };		
			if ( ($i =~ /.+\@.+/) && (!($i =~ /^\</)) && (!($i =~ /^.. /)) && (!($i =~ /^\-/)) ) {
				$tmp = $i; chop($tmp);
				# remove percent signs from email addys
				# this works around a bug in Unix::Syslog
				$tmp =~ s/\%/\*/g;
				push @to, $tmp;
			};
			$i = <HEADER>;
		} while (!($i eq "\n"));
		
		while(<HEADER>) {
			# grab the subject if we find one
			if ($_ =~ /Subject\: /) {
				$tmp = $_; chop($tmp);
				($dummy,$subject) = split /Subject\: /, $tmp, 2;
			};
			
			# kick out old exiscan X-Scanner lines
			next if ($_ =~ /082  X-Scanner\: exiscan/);

			$_ =~ s/^.....//g if ($_ =~ /^[0-9][0-9][0-9]/);
			push @completemail, $_;
		};
		close(HEADER);

		# remove percent signs from email addys
		# this works around a bug in Unix::Syslog
		$from =~ s/\%/\*/g;

	       # ------------------------------------
	       # make sure the @to array does not get too large
	       # this works around a buffer overflow which obviously 
	       # appears when Unix::Syslog is logging to older syslogds
	       # right now we set the limit to 5 addresses plus a
	       # "truncated" message

		if ($#to > 5) {
			$#to = 5;
			push @to, "<more recipients, but not shown>";
		};	

	       # -------------------------------------
	       # do not scan the mail if it is sent to
	       # the postmaster address ONLY

		if ( ($#to == 0) && (@to[0] eq $postmaster) ) {
			print "$id: postmaster-only mail\n" if ($debug);
			syslog LOG_INFO, "ID $id R:PostMaster mail only, dequeuing.";
			
			# dequeue
			$checked{$id} = 1;
                        system("$exim -Mc $id &");
		
			# process next mail
			print "$id: leaving process loop\n" if ($debug);
			next;
		};

	       # -------------------------------------
	       # add our tag to the header
		
		$crypttag = crypt($id,$salt);
		# cut away the salt.
		$crypttag =~ s/^..//;
		push @completemail, "X-Scanner: exiscan \*$id\*$crypttag\* http://duncanthrax.net/exiscan/\n";

		# add separator
		push @completemail, "\n";

	       # -------------------------------------
	       # append the complete body

		open(BODY,"< $queuedir/$id-D");
                $i = <BODY>;	# kick out 1st line
		push @completemail, <BODY>;
                close(BODY);

	      # -------------------------------------
	      # write the complete mail to disk.
	      # that gives us a file in standard MBOX format

		open(TMPMAIL,"> $checkdir/$id-complete");
		print TMPMAIL @completemail;
		close(TMPMAIL);

               # -------------------------------------
               # unpack mail in temporary directory

		mkdir "$checkdir/$id-tmp", 0755;
		
		print "$id: unpacking in $checkdir/$id-tmp\n" if ($debug);
		# use reformime if present
		if ($reformime) {
			chdir("$checkdir/$id-tmp");
			system("$reformime -xf- < $checkdir/$id-complete");
			chdir("$basepath");
		}
		# else we use good 'ol metamail
		else {
			system("METAMAIL_TMPDIR=$checkdir/$id-tmp; export METAMAIL_TMPDIR; $metamail -r -q -x -w $checkdir/$id-complete > /dev/null 2> /dev/null");
		};

	       # -------------------------------------
	       # enumerate the unpacked files and feed
	       # winmail.dat (MS-TNEF) to tnef, if configured

		if ($tnef) {
			opendir(CHECKDIR,"$checkdir/$id-tmp");
			@mimefiles = grep /winmail/i, readdir CHECKDIR;
			closedir(CHECKDIR);

			print "$id: scanned $checkdir/$id-tmp and found TNEF: @mimefiles\n" if ($debug);		

			foreach $mimefile (@mimefiles) {
				system("$tnef -f $checkdir/$id-tmp/$mimefile -C $checkdir/$id-tmp --overwrite");
			};
		};

	       # -------------------------------------
	       # attempt to unpack SMIME encapsulated messages
	       # these will typically come out of MS-TNEF encapsulated messages

		if ($unpack_smime) {
			opendir(CHECKDIR,"$checkdir/$id-tmp");
			@mimefiles = grep /smime/i, readdir CHECKDIR;
			closedir(CHECKDIR);

			print "$id: scanned $checkdir/$id-tmp and found SMIME: @mimefiles\n" if ($debug);
                
			foreach $mimefile (@mimefiles) {
				# use reformime if present
				if ($reformime) {
	                       		chdir("$checkdir/$id-tmp");
			       		system("$reformime -xf- < $checkdir/$id-tmp/$mimefile");
		  			chdir("$basepath");  
                		}
		 		# else we use good 'ol metamail
                		else {
                        		system("METAMAIL_TMPDIR=$checkdir/$id-tmp; export METAMAIL_TMPDIR; $metamail -r -q -x -w $checkdir/$id-tmp/$mimefile > /dev/null 2> /dev/null");
				};
	        	};
		};

               # -------------------------------------
               # run virus scanner and get result
               # save the scanner output in $checkdir/$id-out
	       # each scanner needs to be treated separately
	       # of course. Sigh.

		$result = 0;

		# McAfee's uvscan
		if ($scanner =~ /mcafee/i) {
			$result = system("$scannerex --unzip -r $checkdir/$id-tmp > $checkdir/$id-out 2>&1");
		};

		# Trend Micro's vscan
		if ($scanner =~ /trendmicro/i) {
			open(OUTPUT,"> $checkdir/$id-out");
			open(VSCAN,"$scannerex $checkdir/$id-tmp |");

			while(<VSCAN>) {
				print OUTPUT $_;
			        if ($_ =~ /Infected \: /) {
		        	        chop;
		                	($dummy,$infected) = split / \: /,$_,2;
			                if ($infected =~ /[1-9]/) {
		        	                $result = 1;
        		        	        last;
		        	        };
			        };
			};
			close(VSCAN);
			close(OUTPUT);
		};

		# Sophos Sweep
		if ($scanner =~ /sophos/i) {
			$result = system("$scannerex sweep -archive -f -all -rec $checkdir/$id-tmp > $checkdir/$id-out 2>&1");
		};

		# Kapersky's AVP
		if ($scanner =~ /avp/i) {
			open(OUTPUT,"> $checkdir/$id-out");
			open(VSCAN,"$scannerex $checkdir/$id-complete |");
			
			while(<VSCAN>) {
				print OUTPUT $_;
				if ($_ =~ /Return code/) {
					chop;
					($dummy,$infected) = split /\: /,$_,2;
					if ($infected =~ /[1-8]/) {
						$result = 1;
						last;
					};
				};
			};
			close(VSCAN);
			close(OUTPUT);
		};


		# alternative AVP code (called demoavp)
		# by Sergey Kovalyov <sergey@metaphoria.net>
		if ($scanner =~ /demoavp/i) {
			$infected = system("$scannerex $checkdir/$id-tmp -y > $checkdir/$id-out 2>&1");
                        print "$id: AVP returned $infected\n" if ($debug);
                        
			$infected = int($infected / 256);
                        if (($infected == 3) || ($infected == 4)) {
                                $result = 1;
                        };
		};

		
		# RAV Antivirus
		if ($scanner =~ /rav/i) {
			open(OUTPUT,"> $checkdir/$id-out");
			open(VSCAN,"$scannerex $checkdir/$id-tmp -unzip |");                                                                                                  
                   	while(<VSCAN>) {
				print OUTPUT $_;                                                                                                                                                     
                       		if ($_ =~ /Different virus bodies\: /) {                                                                                                                                    
                          		chop;
					if ($_ =~ /([a-z]):\ ([0-9])/) {
						$result = $2;                                                                                                                             
                           		}                                                                                                                                
                       		}                                                                                                                                                 
                   	}                                                                                                                                                     
                   	close(VSCAN);
			close(OUTPUT);                                               
               }



		if ($result) {
                
		       # ------------------------------------
                       # Virus found
			
			print "$id: virus found\n" if ($debug);

			syslog LOG_INFO, "ID $id F:$from T:@to R:virus found, moving to $virusdir";
	
			move("$queuedir/$id-H","$virusdir/$id-H");
			move("$queuedir/$id-D","$virusdir/$id-D");


		       # ------------------------------------
		       # send notification to admin. Infected mail is attached.
	
			syslog LOG_INFO, "ID $id sending mail attachment to $postmaster";
			system("$metasend -t $postmaster -F exiscan -s \"Virus in mail $id\"  -S 99999999 -b -m text/plain -f $checkdir/$id-out -D 'Output from scanner' -e 7bit -n -m message/rfc822 -f $checkdir/$id-complete -D 'Original Message' -e 7bit");

			# -----------------------------------
			# send notification to sender (if configured)

			if ($sender_notification) {
				if ( (length($from) > 4) && (!($checked{$from})) ) {
					syslog LOG_INFO, "ID $id sending notification to sender ($from)";
					my $cleanfrom = $from; $cleanfrom =~ s/\<|\>//g;
					open(SENDER,"| $mail -s \"Virus\ /\ harmful\ content\ found\ in\ EMail\" $cleanfrom"); 
					print SENDER eval $sender_notification_text;
					print SENDER $notification_footer;
					close(SENDER);
				
					# remember sender notification
					$checked{$from} = 1;
				};
			};


			# -----------------------------------
			# send notification to recipients (if configured)

			if ($rcpt_notification) {
				syslog LOG_INFO, "ID $id sending notification to recipients (@to)";
				open(RCPT,"| $mail -s \"Virus\ /\ harmful\ content\ found\ in\ EMail\" @to"); 
				print RCPT eval $rcpt_notification_text;
				print RCPT $notification_footer;
				close(RCPT);
			};
		}
		else {
		
		       # ------------------------------------
		       # insert X-Scanner header line	

			$crypttag = crypt($id,$salt);
			# cut away the salt.
			$crypttag =~ s/^..//;

			if (-e "$queuedir/$id-H") {
				open(HEADER,">> $queuedir/$id-H");
				print HEADER "082  X-Scanner: exiscan \*$id\*$crypttag\* http://duncanthrax.net/exiscan/\n";
				close(HEADER);
			};

                       # ------------------------------------
                       # remember mail id

			$checked{$id} = 1;

                       # ------------------------------------
                       # Mail is ok, de-queue

			print "$id: clean, launching exim dequeue process\n" if ($debug);
			syslog LOG_INFO, "ID $id F:$from T:@to R:mail clean, dequeuing"; 
			system("$exim -Mc $id &");
		};

               # -------------------------------------
               # clean up temp files

		print "$id: clean up temp files/dirs\n" if ($debug);
		remove \1, "$checkdir/$id-tmp";
		unlink "$checkdir/$id-complete";
		unlink "$checkdir/$id-out";
		
		print "$id: leaving process loop\n" if ($debug);
	};
	print "$id: already worked on it since last hash reset, processing next\n" if ($debug);
   };
 };
 };

# -----------------------------------
# if $quit is not 1 then this code got 
# reached when the superloop eval block died 
# unexpectedly. we try to output some meaningful
# error message to the log and simply restart
# the superloop ..

if (!($quit)) {
	print "$id: error ($@). Last mail ID was $id.\n" if ($debug);
	syslog LOG_INFO, "error occured ($@). Last mail ID was $id.";
	goto SUPERLOOP;
};

# -----------------------------------
# the exit routine. gets passed
# when we shut down with SIGHUP recvd

EXITEXISCAN:

print "$id: terminating with HUP signal\n" if ($debug);
syslog LOG_INFO, "terminating";
closelog;
exit(0);
